
package SI;

# Speckle Interferometry package

use fields qw(acc acc2 accKT df dfn n noise dim accBS dfBS dfnBS);

use PDL;
use PDL::FFTW;
use PDL::Image2D;
use PDL::ImageND;
use PDL::CallExt;
use Utils;
use PDL::Exporter;

@ISA = ('Exporter');
@EXPORT = qw(restoreKT restoreBS normalize APimprove APsupport);

use strict;

sub new {
  my $class = shift;

  no strict 'refs';
  my $self = bless [\%{"$class\::FIELDS"}], $class;
  use strict 'refs';

  $self->{dim} = shift;
  $self->{noise} = shift;
# Note: $self->{noise} is (the variance gaussian noise per pixel * no. of pixels)

  $self->{acc} = zeroes float,$self->{dim},$self->{dim};
  $self->{acc2} = zeroes float,$self->{dim},$self->{dim};
  $self->{n} = 0;

  $self;
}

sub initKT {
  my $self = shift;
  $self->{df} = [@_];
  $self->{dfn} = scalar @_;
  my ($i);
  
  $self->{accKT} = [1 .. $self->{dfn}];
  for ($i=0;$i<$self->{dfn};$i++) {
    $self->{accKT}->[$i] = zeroes float,2,$self->{dim},$self->{dim};
  }
}

sub initBS {
  my $self = shift;
  $self->{dfBS} = [@_];
  $self->{dfnBS} = scalar @_;
  my ($i);
  
  $self->{accBS} = [1 ..$self->{dfnBS}];
  for ($i=0;$i<$self->{dfnBS};$i++) {
    $self->{accBS}->[$i] = zeroes float,2,$self->{dim},$self->{dim};
  }
}

sub add {
    my $self = shift;
    my $im = shift;
    my $rfft = shift;
    my ($xn,$yn) = $im->dims;
    my ($mrfft);

    if (not defined $rfft) {
	$rfft = zeroes float,2,$xn,$yn;
	$rfft->slice("(0),:,:") .= $im;
	nfftw $rfft;
# controlla normalizzazione!
    }
    $mrfft = Cmod2 $rfft;
     if (defined $self->{noise}) {
      
      $mrfft -= sum $im;
      $mrfft -= $self->{noise}; 
    }
    $self->{acc} += $mrfft;
    $self->{acc2} += $mrfft * $mrfft;
    $self->{n} ++;
}

sub addKT {
  my $self = shift;
  my $im = shift;
  my ($i,$dx,$dy);
  my ($xn,$yn) = $im->dims;
  my ($imc,$shifted);

  $imc = zeroes float,2,$xn,$yn;
  $imc->slice("(0),:,:") .= $im;
  nfftw $imc;
# controlla normalizzazione!

  $self->add($im,$imc);
  for ($i=0;$i<$self->{dfn};$i++) {
      $dx = $self->{df}->[$i]->[0];
      $dy = $self->{df}->[$i]->[1];
      $shifted = shift2d($imc->mv(0,2),$dx,$dy)->mv(2,0);
      $self->{accKT}->[$i] += Cmul($imc,Cconj($shifted));
      if (defined $self->{noise})  {
	  if ($dx<0) {$dx = $xn+$dx;}
	  if ($dy<0) {$dy = $yn+$dy;}
	  my $cv = Cconj($imc->slice(":,($dx),($dy)"));
	  $self->{accKT}->[$i] -= $cv->dummy(-1,$xn)->dummy(-1,$yn);
      }
  }
}

sub addBS {
  my $self = shift;
  my $im = shift;
  my ($i,$dx,$dy);
  my ($xn,$yn) = $im->dims;
  my ($imc,$shifted,$fulled);

  $imc = zeroes float,2,$xn,$yn;
  $imc->slice("(0),:,:") .= $im;
  nfftw $imc;
# controlla normalizzazione!

  $self->add($im,$imc);
  for ($i=0;$i<$self->{dfnBS};$i++) {
      $dx = $self->{dfBS}->[$i]->[0];
      $dy = $self->{dfBS}->[$i]->[1];
      $shifted = shift2d($imc->mv(0,2),$dx,$dy)->mv(2,0);
      if ($dx<0) {$dx = $xn+$dx;}
      if ($dy<0) {$dy = $yn+$dy;}
      my $cv = $imc->slice(":,($dx),($dy)");
      $fulled = $cv->dummy(-1,$xn)->dummy(-1,$yn);
      $self->{accBS}->[$i] += Cmul(Cmul($imc,$fulled),Cconj($shifted));
      if (defined $self->{noise})  {
	  my $sim = $im->sum;
	  $sim = float [$sim,$sim];
	  $sim = $sim->dummy(-1,$xn)->dummy(-1,$yn);
	  my $snoise = $self->{noise};
	  $snoise = float [$snoise,$snoise];
	  $snoise = $snoise->dummy(-1,$xn)->dummy(-1,$yn);
	  $self->{accBS}->[$i] += 2*$sim + 3*$snoise
	      - $imc->Cmod2->dummy(0,2)
		  - $fulled->Cmod2->dummy(0,2) 
		      - $shifted->Cmod2->dummy(0,2);
      }
  }
}

sub get {
  my $self = shift;
  
  return $self->{acc}/$self->{n};
}

sub getVar {
  my $self = shift;
  
  return $self->{acc2}/$self->{n} - $self->get*$self->get;
}

sub getSNR {
  my $self = shift;
  my $thr = max($self->getVar)/1e23;
  my $var = $self->getVar;

  $var = ($var>$thr)*$var + ($var<=$thr)*$thr;
  return $self->get / sqrt($var)*float(sqrt($self->{n}));
}

sub getKT {
  my $self = shift;
  my $i = shift;
   
  return $self->{accKT}->[$i] / $self->{n};
}

sub getBS {
  my $self = shift;
  my $i = shift;
   
  return $self->{accBS}->[$i] / $self->{n};
}

sub normalize {
  my ($in) = @_;
  my $tmp;

  $tmp = Cmod $in ;
  $tmp = $tmp * ($tmp!=0.0) + float(1e30) * ($tmp==0.0);
  return $in / $tmp->dummy(0,2);
}

sub restoreKT {
  my ($mod,$modr,$snr,$dxrp,$dxrn,$dyrp,$dyrn,$maxf,$alpha,$thr) = @_;
  my $dim = ($mod->dims) [1];
  my $phase = zeroes float,$dim,$dim;
  my ($i,$j,$t,$mask,$modratio,$mask2);
  my $image = zeroes float,2,$dim,$dim;
  my $minthr;
  my ($ophase, $omod);
  my $dir = $ENV{'SEEINGLIB'};

  $dxrp = normalize($dxrp);
  $dxrn = normalize($dxrn);
  $dyrp = normalize($dyrp);
  $dyrn = normalize($dyrn);

  callext("${dir}/c/KT.so",'phase1',$dxrp,$dxrn,$dyrp,$dyrn,$image);

  $mod *= 1000/max($mod);
  $modr *= 1000/max($modr);
  $minthr = max($mod) * $thr;
  $mod = $mod * ($mod>$minthr) + $minthr * ($mod<=$minthr);
  $minthr = max($modr) * $thr;
  $mod = $mod * ($modr>$minthr);
  $modr = $modr * ($modr>$minthr) + $minthr * ($modr<=$minthr);
  $snr = $snr * ($snr>0.001) + 0.001 * ($snr<=0.001);
  $snr = shift2d($snr,$dim/2,$dim/2);
  $snr = circ_mean $snr,{Center=>[$dim/2,$dim/2]};
  $snr = shift2d($snr,-$dim/2,-$dim/2);
  $modr = shift2d($modr,$dim/2,$dim/2);
  $modr = circ_mean $modr,{Center=>[$dim/2,$dim/2]};
  $modr = shift2d($modr,-$dim/2,-$dim/2);
  $modratio = sqrt( $mod / ($modr + $alpha / $snr ) );
  $image *= $modratio->dummy(0,2);

  $mask = rvals( float,$dim,$dim,{Centre=>[($dim-1)/2,($dim-1)/2]} );
  $mask = shift2d($mask,$dim/2,$dim/2);
  $mask = $mask <= $maxf;
  $image *= $mask->dummy(0,2);

  $ophase = Carg $image;
  $omod = Cmod $image;

  $image = Cmod ifftw  $image;
  $image /= max $image;

  return $image,$ophase,$omod,$snr;
}

sub restoreBS {
  my ($mod,$modr,$snr,$dxrp,$dxrn,$dyrp,$dyrn,$maxf,$alpha,$thr) = @_;
  my $dim = ($mod->dims) [1];
  my ($i,$j,$t,$mask,$modratio,$mask2);
  my $image = zeroes float,2,$dim,$dim;
  my $minthr;
  my ($ophase, $omod);
  my $dir = $ENV{'SEEINGLIB'};

  $dxrp = normalize($dxrp);
  $dxrn = normalize($dxrn);
  $dyrp = normalize($dyrp);
  $dyrn = normalize($dyrn);

  callext("${dir}/c/KT.so",'phase2',$dxrp,$dxrn,$dyrp,$dyrn,$image);

  # modulus
  $mod *= 1000/max($mod);
  $modr *= 1000/max($modr);
  $minthr = max($mod) * $thr;
  $mod = $mod * ($mod>$minthr) + $minthr * ($mod<=$minthr);
  $minthr = max($modr) * $thr;
  $mod = $mod * ($modr>$minthr);
  $modr = $modr * ($modr>$minthr) + $minthr * ($modr<=$minthr);
  $snr = $snr * ($snr>0.001) + 0.001 * ($snr<=0.001);
  $snr = shift2d($snr,$dim/2,$dim/2);
  $snr = circ_mean $snr,{Center=>[$dim/2,$dim/2]};
  $snr = shift2d($snr,-$dim/2,-$dim/2);
  $modr = shift2d($modr,$dim/2,$dim/2);
  $modr = circ_mean $modr,{Center=>[$dim/2,$dim/2]};
  $modr = shift2d($modr,-$dim/2,-$dim/2);
  $modratio = sqrt( $mod / ($modr + $alpha / $snr ) );
  $image *= $modratio->dummy(0,2);

  # top hat filter
  $mask = rvals( float,$dim,$dim,{Centre=>[($dim-1)/2,($dim-1)/2]} );
  $mask = shift2d($mask,$dim/2,$dim/2);
  $mask = $mask <= $maxf;
  $image *= $mask->dummy(0,2);

  # save info
  $ophase = Carg $image;
  $omod = Cmod $image;

  # ifft
  $image = Cmod ifftw  $image;
  $image /= max $image;

  return $image,$ophase,$omod,$snr;
}

sub APfilter {
  my($snr,$th) = @_;

  return ($snr>$th) + ($snr<=$th)*$snr/$th;
}

sub APsupport {
  my ($im,$th) = @_;
  my ($t,$a,$tim,$tt);
  my ($xn,$yn) = $im->dims;
  
  $tim = zeroes float,2*$xn,2*$yn;
  $tt = $tim->mslice([$xn/2,3*$xn/2-1],[$yn/2,3*$yn/2-1]);
  $tt .= $im;
  $t = rfftw $tim;
  $a = irfftw ( $t * Cconj($t) );
  $a = $a->mslice([$xn/2,3*$xn/2-1],[$yn/2,3*$yn/2-1]);
  $a = shift2d($a,64,64);
  $a /= max $a;
  return $a>$th;
}

sub APimprove {
  my ($image,$snr,$maxiter,$snrt,$th) = @_;
  my ($i,$tg,$g,$lg,$F,$filter,@e,$np,$s);
  my ($xn,$yn) = $image->dims;
  my $n = $xn*$yn;

  if (not defined $th) {$th = -1;}
  
  $g = $image->copy;
  $g /= abs( max($g));
  $lg = $g;

  $F = rfftw $g;
  $filter = APfilter($snr,$snrt);

  if ($th>0) {
    $s = APsupport($image,$th);
  }

  for($i=0;$i<$maxiter;$i++) {
    print "$i of $maxiter\n";
    $tg = rfftw $g;
    # fourier domain
    $tg += $filter->dummy(0,2) * ($F-$tg); 
    $g = irfftw $tg;
    $g /= $n;
    # direct domain
    # positivity 
    $np = sum(($g<0)*$g);
    $g = $g*($g>0);
    $g /= abs( max($g));
    $e[$i] = sqrt(sum(($g-$lg)*($g-$lg)));
    $lg = $g->copy;
  }
  
  return ($g,$filter,@e);
}

1;
