/*
  layer_fft.c

  CopyLeft GPL by Christian Pellegrin 1999. See COPYING.

  Syntax: layer_fft [options] filename
  Options:
  -s int: set scale
  -x double: x dimesion of layer
  -y double: y dimension of layer
  -r double: r0
  -L double: L0
  -h int: number of subharmonics
  -k int: aspect ratio correction 0=none, 1=auto, 2=user supplied
  -w double: x aspect ratio
  -W double: y aspect ratio
  -f string: outputs structure function to file
  -v :be verbose
  -z string: load wisdom from filename
  -e int: set seed
  -m int: subharmonic method (1,2)
  -S int: subharmonic screen reduction ratio
  -c string: read coefficents of subharmonics from file
  -D int: calculate structure function in number iterations
  -G double: use a Gaussian freq filter with given SIGMA
  -A double: SIGMA in direct space (amplitude of initial gaussian noise) (only for -G and -H)
  -a :to keep power integral kostant in gaussian mode (only for -G and -H)
  -H double: top hut freq filter at given spatial frequency 
  -T double: non Taylor case: cuts after given value the filter
*/


#include "config.h"

#include "layer.h"
#include "image.h"

#include <gsl_math.h>
#include <gsl_matrix.h>
#include <gsl_matrix_op.h>
#include <gsl_matrix_applicator.h>
#include <gsl_rng.h>
#include <gsl_randist.h>
#include <gsl_complex.h>
#include <gsl_complex_op.h>
#include <getopt.h>
#include <stdio.h>
#include <time.h>

struct layer_header_fft layer;

int CI,CJ;
REAL F,L0_2;

int verbose=0,totn;
char fname[MAXL];
char sf_fname[MAXL];
int out_sf=0;
gsl_rng * rng;
double stots,stotn;
double expo=11.0/12.0;
int gaussian=0;
int top_hat=0;
double gVar,gSigma;
double gA = 1.0;
double gKeepIt = 0.0;
int gKeep = 0;
double thCut;
int nt=0;
double ntCut;

MATRIX* screen = NULL;
MATRIX* D = NULL;

void remap(MATRIX * out, MATRIX * in, int fac) {
  int in_i, in_j, out_i,out_j,i,j;
  REAL y1,y2,y3,y4,t,u,temp;

  for(in_i=0, out_i=0; in_i<(in->size1-1); in_i++, out_i+=fac) 
    for(in_j=0, out_j=0; in_j<(in->size2-1); in_j++, out_j+=fac) {
      y1=MATRIX_GET(in,in_i,in_j);
      y2=MATRIX_GET(in,in_i+1,in_j);
      y3=MATRIX_GET(in,in_i+1,in_j+1);
      y4=MATRIX_GET(in,in_i,in_j+1);
      for(i=0; i<fac; i++) 
	for(j=0; j<fac; j++) {
	  t= ((REAL) i) / fac;
	  u= ((REAL) j) / fac;
	  temp=MATRIX_GET(out,i+out_i,j+out_j);
	  temp+= (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4;
	  MATRIX_SET(out,i+out_i,j+out_j,temp);
	}
    }
}

void do_gaussian_noise(REAL *s) {
  if (gaussian || top_hat) {
    *s=gsl_ran_gaussian(rng,gA)/(stotn*stotn); 
  }
  else {
    *s=gsl_ran_gaussian(rng,1)/stotn; 
  }
}

void do_subh_noise(COMPLEX *s) {
  GSL_REAL(*s)=gsl_ran_gaussian(rng,1)/stots; 
  GSL_IMAG(*s)=0;
}

void do_filter(COMPLEX *h, int i, int j) {
  REAL f;
  REAL si;
  REAL sj;
  REAL r2;

  if (i<=CI) si= ((REAL) (i) ) / layer.x;
  else si= ((REAL) (layer.xn-i) ) / layer.x; 
  if (j<=CJ) sj= ((REAL) (j) ) / layer.y;
  else {
    fprintf(stderr,"there's a bug!\n");
  }
  r2 = si*si+sj*sj;
  if (gaussian) {
    f= exp ( -r2/2.0/gVar);
    MATRIX_COMPLEX_SCALE(*h,f);
    if (gKeep)
      gKeepIt += f*f;
  }
  else if (top_hat) {
    if ( r2>(thCut*thCut) ) {
      REAL dummy = 0.0;
      MATRIX_COMPLEX_SCALE(*h,dummy);
    }
    else 
      gKeepIt += 1.0;
  }
  else {
    if (nt) {
      if (r2<(1.0/(ntCut*ntCut)))
	f = 0;
      else
	f= F / pow( r2+L0_2, 11.0/12.0 );
    }
    else 
      f= F / pow( r2+L0_2, 11.0/12.0 );
    MATRIX_COMPLEX_SCALE(*h,f);
  }
}

void do_sf(COMPLEX *h, int i, int j) {
  
  GSL_REAL(*h)=pow(-1,i+j);
  GSL_IMAG(*h)=0;

  do_filter(h,i,j);
  do_filter(h,i,j);
}

#define C(i,j) ( (i-layer.mins)*layer.tots + (j-layer.mins) )
#define CC(p,i,j) ( (p-1)*layer.tots*layer.tots + (i-layer.mins)*layer.tots + (j-layer.mins) )

REAL f_lf(int m, int n, int p) {
  double p3,xx,yy,r2;

  if (layer.sub_method==1 && (m==0 && n==0) ) return 0;
  if (layer.sub_method==2 && ( abs(m+0.5) < 0.6 && abs(n+0.5) < 0.6 ) ) return 0.0;

  p3=pow(3.0,-p);
  xx=(p3*(((REAL) m)+layer.offs))/layer.x;
  yy=(p3*(((REAL) n)+layer.offs))/layer.y;
  r2 = xx*xx + yy*yy;
  if (nt && r2<L0_2)
    return 0;
  else
    return F * p3 / pow( r2 + L0_2, expo );
}

void do_sub_screen(COMPLEX_MATRIX **h, MATRIX *screen, float weight[]) {
  int no,mo,i,j,p;
  double cc;
  REAL *fa;
  double CI,CJ,ni,mi;

  fa=malloc(layer.n_sub*layer.tots*layer.tots*sizeof(REAL));

  for(i=layer.mins;i<=layer.maxs;i++)
    for(j=layer.mins;j<=layer.maxs;j++) 
      for(p=1;p<=layer.n_sub;p++) {
	fa[CC(p,i,j)]=f_lf(i,j,p)*weight[p-1];
      }
  CI=screen->size1/2.0;
  CJ=screen->size2/2.0;
  for(no=0;no<screen->size1;no++) 
    for(mo=0;mo<screen->size2;mo++) {
      int p;
      double phase=0;
      
      ni=no-CI;
      mi=mo-CJ;
      for(p=1;p<=layer.n_sub;p++) {
	REAL k;
	
	k=2*M_PI*pow(3.0,-p);
	for(i=layer.mins;i<=layer.maxs;i++)
	  for(j=layer.mins;j<=layer.maxs;j++) {
	    double k1;
	    double c;
	    
	    k1= k * ( ( ((REAL) i) + layer.offs )*ni/screen->size1 + 
		      ( ((REAL) j) + layer.offs )*mi/screen->size2 );
	    c=cos(k1);
	    if (h!=NULL) {
	      double f=fa[CC(p,i,j)];
	      COMPLEX nv = MATRIX_COMPLEX_GET(h[p-1],i-layer.mins,j-layer.mins);

	      c=f*( GSL_REAL(nv)*c - GSL_IMAG(nv)*sin(k1) );
	    } 
	    else {
	      double f=fa[CC(p,i,j)];
	      
	      f*=f;
	      c*=f;
	    }
	    phase+=c;
	  }
      }
      MATRIX_SET(screen,no,mo,phase);
    }

  free(fa);
  
}

#define MAXW 20

int main(int argc, char *argv[]) {
  int opt,seed,i,j,wread=0;
  float weight[MAXW];
  rfftwnd_plan plan_d;
  rfftwnd_plan plan_r;
  FILE *f;
  char wisdom[MAXL];
  int out_wisdom=0;
  COMPLEX_MATRIX cm;
  COMPLEX zero={{0,0}};
  char buff[MAXL];
  int do_out_gray = 0;
  int iterate_D=0,times2repeat=1,rt;

  rng = gsl_rng_alloc (gsl_rng_taus);
  seed=time(NULL);
  layer.sub_method=1;
  layer.n_sub=0;

  for(i=0;i<MAXW;i++) weight[i]=1.0;

  while ( (opt=getopt(argc,argv,"gs:x:y:r:L:h:k:w:W:f:vz:e:m:S:c:D:G:A:aH:T:")) != EOF ) {
    switch(opt) {
    case 'v':
      verbose=1;
      break;
    case 'g':
      do_out_gray=1;
      break;
    case 's':
      sscanf(optarg,"%d",&layer.scale);
      break;
    case 'D':
      sscanf(optarg,"%d",&iterate_D);
      times2repeat = iterate_D;
      break;
    case 'x':
      sscanf(optarg,"%lg",&layer.x);
      break;
    case 'y':
      sscanf(optarg,"%lg",&layer.y);
      break;
    case 'r':
      sscanf(optarg,"%lg",&layer.r0);
      break;
    case 'L':
      sscanf(optarg,"%lg",&layer.L0);
      break;
    case 'h':
      sscanf(optarg,"%d",&layer.n_sub);
      break;      
    case 'k':
      sscanf(optarg,"%d",&layer.correct);
      break;      
    case 'w':
      sscanf(optarg,"%lg",&layer.wx);
      break;      
    case 'W':
      sscanf(optarg,"%lg",&layer.wy);
      break;      
    case 'f':
      strcpy(sf_fname,optarg);
      out_sf=1;
      break;
    case 'z':
      strcpy(wisdom,optarg);
      f=fopen(wisdom,"r");
      if (f!=NULL) {
	fftw_import_wisdom_from_file(f);
	fclose(f);
      }
      else fprintf(stderr,"Cannot Import Wisdom!\n");
      out_wisdom=1;
      break;
    case 'e':
      sscanf(optarg,"%d",&seed);
      break;
    case 'm':
      sscanf(optarg,"%d",&layer.sub_method);
      break;
    case 'S':
      sscanf(optarg,"%d",&layer.red);
      break;
    case 'c':
      wread=1;
      f=fopen(optarg,"r");
      if (!f) {
	fprintf(stderr,"cannot open coefficent file!\n");
      }
      i=0;
      while(!feof(f)) {
	double t=1.0;
	char b[100];

	fgets(b,100,f);
	t=atof(b);
	if (t==0.0) break;
	weight[i]=t;
	i++;
      }
      fclose(f);
      break;
    case 'G':
      gaussian = 1;
      sscanf(optarg,"%lg",&gSigma);
      gVar = gSigma*gSigma;
      break; 
    case 'A':
      sscanf(optarg,"%lg",&gA);
      break;
    case 'a':
      gKeep = 1;
      break;
    case 'H':
      top_hat = 1;
      sscanf(optarg,"%lg",&thCut);
      break;      
    case 'T':
      sscanf(optarg,"%lg",&ntCut);
      nt = 1;
      break;
   default:
      fprintf(stderr,"Unknown option!\n");
      exit(1);
    }
  }
  argv+=optind;

  if ((argc-optind)!=1) {
    fprintf(stderr,"No output filename for layer or erratic syntax!\n");
    exit(1);
  }
  strcpy(fname,argv[0]);

  layer.xn=(int) (layer.x*layer.scale);
  layer.yn=(int) (layer.y*layer.scale);
  if (layer.n_sub>0) {
    layer.sxn=1+(int) (layer.xn/layer.red);
    layer.syn=1+(int) (layer.yn/layer.red);
  }
  totn=layer.xn*layer.yn;
  gsl_rng_set(rng,seed);
  if (layer.sub_method==1) {
    layer.mins=-1;
    layer.maxs=1;
    layer.tots=3;
    layer.offs=0;
  }
  else if (layer.sub_method==2) {
    layer.mins=-3;
    layer.maxs=2;
    layer.tots=6;
    layer.offs=0.5;
  }
  else {
    fprintf(stderr,"Unknown subharmonic method!\n");
    exit(1);
  }
    
  if (layer.correct==1) {
    double ar=layer.x/layer.y;
    
    if (ar<(1.0/6.0)) {layer.wy=0.3; layer.wx=3.7;}
    else if (ar<(1.0/3.0)) {layer.wy=0.5; layer.wx=2.6;}
    else if (ar<(1.0/1.5)) {layer.wy=0.8; layer.wx=1.8;}
    else if (ar<1.5) {layer.wx=1.0; layer.wy=1.0;}
    else if (ar<3.0) {layer.wx=0.8; layer.wy=1.8;}
    else if (ar<6.0) {layer.wx=0.5; layer.wy=2.6;}
    else {layer.wx=0.3; layer.wy=3.7;}
  }
  
  screen=MATRIX_ALLOC(layer.xn,2*(layer.yn/2+1));
  if (screen==NULL) {
    fprintf(stderr,"Cannot Allocate enough memory!\n");
    exit(1);
  }

  plan_d= rfftw2d_create_plan_specific(layer.xn,layer.yn,FFTW_REAL_TO_COMPLEX, FFTW_OPTIONS | FFTW_IN_PLACE,
				       (fftw_real *) screen->data, 1, NULL, 1);
  plan_r= rfftw2d_create_plan_specific(layer.xn,layer.yn,FFTW_COMPLEX_TO_REAL, FFTW_OPTIONS | FFTW_IN_PLACE,
				       (fftw_real *) screen->data, 1, NULL, 1);

  CI=layer.xn/2;
  CJ=layer.yn/2;
  F=0.15132 / ( sqrt(layer.x*layer.y) * pow(layer.r0,5.0/6.0) );
  L0_2= 1/(layer.L0*layer.L0);    
  stotn=sqrt(totn);
  if (layer.sub_method==1) stots=3*(1+1.0/9.0); 
  if (layer.sub_method==2) stots=6*(1+1.0/9.0);

  if (iterate_D>0) {
    D = MATRIX_ALLOC(layer.xn,layer.yn);
    for(i=0;i<layer.xn;i++)
      for(j=0;j<layer.yn;j++)
	MATRIX_SET(D,i,j,0.0);
  }
  
  for(rt=0;rt<times2repeat;rt++) {
    
    if (verbose && rt==0) {
      fprintf(stderr,"Starting layer creation via FFT & SubHarmonics\n");
      fprintf(stderr,"Output filename: %s\n",fname);
      fprintf(stderr,"Image scale: %d\n",layer.scale);
      fprintf(stderr,"x: %g\n",layer.x);
      fprintf(stderr,"y: %g\n",layer.y);
      fprintf(stderr,"xn: %d\n",layer.xn);
      fprintf(stderr,"yn: %d\n",layer.yn);
      if (!gaussian) {
	fprintf(stderr,"r0: %g\n",layer.r0);
	fprintf(stderr,"L0: %g\n",layer.L0);
      }
      fprintf(stderr,"N subharmonics: %d\n",layer.n_sub);
      if (layer.n_sub>0) {
	fprintf(stderr,"Subharmonics method: %d\n",layer.sub_method);
	fprintf(stderr,"Subharmonic scrren xn: %d\n",layer.sxn);
	fprintf(stderr,"Subharmonic scrren yn: %d\n",layer.syn);
      }
      if (layer.correct==1) {
	fprintf(stderr,"Automatic aspect ratio correction\n");
	fprintf(stderr,"Wx: %g\n",layer.wx);
	fprintf(stderr,"Wy: %g\n",layer.wy);
      }
      if (layer.correct==2) {
	fprintf(stderr,"Wx: %g\n",layer.wx);
	fprintf(stderr,"Wy: %g\n",layer.wy);
      }
      if (out_sf==1) {
	fprintf(stderr,"Structure function will be outputed to: %s\n",sf_fname);
      }
      if (wread) {
	fprintf(stderr,"Subharmonics weights:\n");
	for(i=0;i<layer.n_sub;i++) fprintf(stderr,"%d: %g\n",i,weight[i]);
      }
      if (gaussian) {
	fprintf(stderr,"Running in Gaussian mode::\n");
	fprintf(stderr,"Variance  in frequrncy gaussian filter : %g (sigma %g)\n",gVar,gSigma);
	fprintf(stderr,"Sigma (circa direct space amplitude): %g\n",gA);
	if (gKeep) 
	  fprintf(stderr,"Keep power costant between domains\n");
	else
	  fprintf(stderr,"Don't keep power costant between domains\n");
      }
      if (nt)
	fprintf(stderr,"Non Taylor mode with cut: %g\n",ntCut);
    }

    MATRIX_APPLY(screen,do_gaussian_noise);
    rfftwnd_one_real_to_complex(plan_d,(fftw_real *) screen->data,NULL);
    cm.size1= layer.xn;
    cm.size2= layer.yn/2+1;
    cm.data= screen->data;
    MATRIX_COMPLEX_APPLY_IJ(&cm,do_filter);
    if ((gaussian || top_hat) && gKeep) {
      double korr = sqrt(stotn*(stotn/2+1)/gKeepIt);
      /*      fprintf(stderr,"korr is %g\n",korr); */
      MATRIX_COMPLEX_MUL_SCALAR(&cm,korr);
    }
    MATRIX_COMPLEX_SET(&cm,0,0,zero);
    if (layer.n_sub>0 && layer.sub_method==2) {
      COMPLEX t;
    
      t=MATRIX_COMPLEX_GET(&cm,1,0);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,1,0,t);

      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,0);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,0,t);

      t=MATRIX_COMPLEX_GET(&cm,0,1);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,0,1,t);

      t=MATRIX_COMPLEX_GET(&cm,1,1);
      MATRIX_COMPLEX_SCALE(t,0.75);
      MATRIX_COMPLEX_SET(&cm,1,1,t);

      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,1);
      MATRIX_COMPLEX_SCALE(t,0.75);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,1,t);
    }
    if (layer.correct!=0) {
      COMPLEX t;
    
      t=MATRIX_COMPLEX_GET(&cm,0,1);
      MATRIX_COMPLEX_SCALE(t,layer.wy);
      MATRIX_COMPLEX_SET(&cm,0,1,t);
    
      t=MATRIX_COMPLEX_GET(&cm,1,0);
      MATRIX_COMPLEX_SCALE(t,layer.wx);
      MATRIX_COMPLEX_SET(&cm,1,0,t);
    
      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,0);
      MATRIX_COMPLEX_SCALE(t,layer.wx);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,0,t);
    }

    rfftwnd_one_complex_to_real(plan_r,(fftw_complex *) screen->data,NULL);  
    correct_rfftw(screen);
  
    if (layer.n_sub>0) {
      COMPLEX_MATRIX **noise;
      MATRIX *subscreen = MATRIX_CALLOC(layer.sxn,layer.syn);
      fftwnd_plan s_plan_d;
      int i;

      noise=malloc(layer.n_sub*sizeof(COMPLEX_MATRIX*));

      for(i=0;i<layer.n_sub;i++)
	noise[i]=MATRIX_COMPLEX_CALLOC(layer.tots,layer.tots);

      s_plan_d= fftw2d_create_plan_specific(layer.tots,layer.tots,FFTW_FORWARD,
					    FFTW_OPTIONS | FFTW_IN_PLACE, 
					    (fftw_complex *) noise[0]->data, 1, NULL, 1);

      for(i=0;i<layer.n_sub;i++) {
	MATRIX_COMPLEX_APPLY(noise[i],do_subh_noise);
	fftwnd_one(s_plan_d,(fftw_complex *) noise[i]->data,NULL);
      }

      do_sub_screen(noise,subscreen,weight);
      remap(screen,subscreen,layer.red);
      MATRIX_FREE(subscreen);
      free(noise);
    }

    if (iterate_D==0) {
      if ((f=fopen(fname,"w"))) {
	strcpy(layer.desc,"layer");
	out_data(f,screen);
	fclose(f);
      }
      else {
	fprintf(stderr,"Cannot wirite to file!\n");
	exit(1);
      }
      sprintf(buff,"%s.hdr",fname);
      if ((f=fopen(buff,"w"))) {
	fprintf(f,"%u\n 2\n %u %u\n",DATA_TYPE,layer.yn,layer.xn);
	fclose(f); }
      else {
	fprintf(stderr,"Cannot wirite to file!\n");
	exit(1);
      }
  
      if (do_out_gray) {
	sprintf(buff,"%s.gray",fname);
	if ((f=fopen(buff,"w"))) {
	  out_gray(f,screen,0);
	  fclose(f);
	}
	else {
	  fprintf(stderr,"Cannot wirite to file!\n");
	  exit(1);
	}
      }
    }

    if (iterate_D>0) {
      double mean,t;

      mean = (MATRIX_GET(screen,layer.xn/2-1,layer.yn/2-1) +
	MATRIX_GET(screen,layer.xn/2,layer.yn/2-1) +
	MATRIX_GET(screen,layer.xn/2-1,layer.yn/2) +
	MATRIX_GET(screen,layer.xn/2,layer.yn/2) ) / 4.0;
      for(i=0;i<layer.xn;i++)
	for(j=0;j<layer.yn;j++) {
	  t = MATRIX_GET(screen,i,j) - mean;
	  t *= t;
	  MATRIX_SET(D,i,j,MATRIX_GET(D,i,j) + t);
	}
    }
  }

  if (iterate_D>0) {
      for(i=0;i<layer.xn;i++)
	for(j=0;j<layer.yn;j++) {
	  MATRIX_SET(D,i,j,MATRIX_GET(D,i,j)/times2repeat);
	}
      if ((f=fopen(fname,"w"))) {
	strcpy(layer.desc,"layer");
	fwrite(D->data,sizeof(REAL),layer.xn*layer.yn,f);
	fclose(f);
      }
      else {
	fprintf(stderr,"Cannot wirite to file!\n");
	exit(1);
      }
      sprintf(buff,"%s.hdr",fname);
      if ((f=fopen(buff,"w"))) {
	fprintf(f,"%u\n 2\n %u %u\n",DATA_TYPE,layer.yn,layer.xn);
	fclose(f); }
      else {
	fprintf(stderr,"Cannot wirite to file!\n");
	exit(1);
      }
  
      if (do_out_gray) {
	sprintf(buff,"%s.gray",fname);
	if ((f=fopen(buff,"w"))) {
	  out_gray(f,D,0);
	  fclose(f);
	}
	else {
	  fprintf(stderr,"Cannot wirite to file!\n");
	  exit(1);
	}
      }
  }
  
  if (out_sf && iterate_D==0) {
    MATRIX_COMPLEX_APPLY_IJ(&cm,do_sf);
    if (layer.n_sub>0 && layer.sub_method==2) {
      COMPLEX t;

      t=MATRIX_COMPLEX_GET(&cm,1,0);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,1,0,t);

      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,0);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,0,t);
      
      t=MATRIX_COMPLEX_GET(&cm,0,1);
      MATRIX_COMPLEX_SCALE(t,0.5);
      MATRIX_COMPLEX_SET(&cm,0,1,t);

      t=MATRIX_COMPLEX_GET(&cm,1,1);
      MATRIX_COMPLEX_SCALE(t,0.75);
      MATRIX_COMPLEX_SET(&cm,1,1,t);

      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,1);
      MATRIX_COMPLEX_SCALE(t,0.75);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,1,t);
    }
    if (layer.correct!=0) {
      COMPLEX t;
      
      t=MATRIX_COMPLEX_GET(&cm,0,1);
      MATRIX_COMPLEX_SCALE(t,layer.wy);
      MATRIX_COMPLEX_SET(&cm,0,1,t);
      
      t=MATRIX_COMPLEX_GET(&cm,1,0);
      MATRIX_COMPLEX_SCALE(t,layer.wx);
      MATRIX_COMPLEX_SET(&cm,1,0,t);

      t=MATRIX_COMPLEX_GET(&cm,cm.size1-1,0);
      MATRIX_COMPLEX_SCALE(t,layer.wx);
      MATRIX_COMPLEX_SET(&cm,cm.size1-1,0,t);

    }
    rfftwnd_one_complex_to_real(plan_r,(fftw_complex *) screen->data,NULL);  
    correct_rfftw(screen);

    if (layer.n_sub>0) {
      MATRIX *subscreen=MATRIX_CALLOC(layer.sxn,layer.syn);

      do_sub_screen(NULL,subscreen,weight);
      remap(screen,subscreen,layer.red);
      MATRIX_FREE(subscreen);
    }

    if ((f=fopen(sf_fname,"w"))) {
      strcpy(layer.desc,"structure function");
      out_data(f,screen);
      fclose(f);
    }
    else {
      fprintf(stderr,"Cannot wirite to file!\n");
      exit(1);
    }
    sprintf(buff,"%s.hdr",sf_fname);
    if ((f=fopen(buff,"w"))) {
      fprintf(f,"4\n 2\n %u %u\n",layer.xn,layer.yn);
      fclose(f); }
    else {
      fprintf(stderr,"Cannot wirite to file!\n");
      exit(1);
    }
    
    if (do_out_gray) {
      sprintf(buff,"%s.gray",sf_fname);
      if ((f=fopen(buff,"w"))) {
	out_gray(f,screen,0);
	fclose(f);
      }
      else {
	fprintf(stderr,"Cannot wirite to file!\n");
	exit(1);
      }
    }
  }

  if (out_wisdom) {
    f=fopen(wisdom,"w");
    if (f!=NULL) {
      fftw_export_wisdom_to_file(f);
      fclose(f);
    }
  }
  
  return 0;
  
}
